#
# 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#

### \author[Artem V. Andreev]
### \copyright[2001, 2002 Artem V. Andreev --- See \ref[sl(1)] for details]

### \section[NAME]
#### \em[DocStrip] -- a documentation extractor

### \section[OVERVIEW]
### \em[DocStrip] extracts documentation embedded into C-like sources or scripts
### and outputs it in a given format.
###
### It can handle two different sets of doc per single file -- one for ordinary users
### and another for developers
###
### \em[DocStrip] uses its own macro language to format docs. 
###
### \em[DocStrip] is written in \em[Tensile] and is part of its distribution
### (see \ref[sl(1)])
.ifhas ::module:time
.module <time>
.endif
.module <files>
.include <getopt.sci>

.do { setenv(SCRIPT_PATH getenv(SCRIPT_PATH)\:sysvar(env:location)scripts/docstrip) }

.automaton docstrip
.define true -1;
.define Process 1;
.define StartDoc 2;
.define StartLine 3;
.define InError 4;
.define DocStrip 5;
.define first_slash 100;
.define second_slash 101;
.define first_hash 102;
.define second_asterisk 103;
.define third_asterisk 104;
.define second_hash 105;
.define to_eol 200;
.define startdoc1 201;
.define startdoc2 202;
.define startdoc2_5 203;
.define startline1 300;
.define startline2 301;
.define maindoc 400;
.define after_space 401;
.define after_slash 402;
.define after_asterisk 403;
.define skip_comment 404;
.define skip_comment1 405;
.define skip_spaces 406;
.start Process;
.final InError;
.final DocStrip;

.range ID ['A'-'Z', 'a'-'z', '0'-'9', '_'];
.range ID1 ['A'-'Z', 'a'-'z', '_'];
Process; $$ = '/'; true; first_slash; {}{}
Process; $$ = '#'; true; first_hash; {}{}
Process; true; true; Process; {}{}
first_slash; $$ = '/'; true; second_slash; {}{}
first_slash; $$ = '*'; true; second_asterisk; {}{}
first_slash; true; true; to_eol; {}{}
first_hash; $$ = '#'; true; second_hash; {}{}
first_hash; true; true; to_eol; {}{}
second_slash; ($$:2) = '/'; true; StartDoc; {}{}
second_slash; true; true; to_eol; {}{}
second_asterisk; $$ = '*'; true; third_asterisk; {}{}
second_asterisk; true; true; skip_comment; {}{}
third_asterisk; ($$:2) = '*'; true; StartDoc; {} {}
third_asterisk; true; true; skip_comment; {}{}
second_hash; ($$:2) = '#'; true; StartDoc; {} {}
second_hash; true; true; to_eol; {}{}

to_eol; $$ = '\n'; true; DocStrip; {}{}
to_eol; true; true; to_eol; {}{}
skip_comment; $$ = '*'; true; skip_comment1; {}{}
skip_comment; true; true; skip_comment; {}{}
skip_comment1; $$ = '/'; true; to_eol; {}{}
skip_comment1; true; true; skip_comment; {}{}

StartDoc; $$ = $2; true; StartDoc; {} {}
StartDoc; $$ = '/'; $2 = '*'; to_eol; {}{}
StartDoc; $$ = '\n'; true; StartDoc; {} {}
StartDoc; ^$$ = '\\'; true; maindoc; {}{}
StartDoc; ($$ = ' ') || ($$ = '\t'); true; startdoc1; {} {}
StartDoc; ^true; true; startdoc2; { "\\subsection[" } {}

startdoc1; ($$ = ' ') || ($$ = '\t'); true; startdoc1; {} {}
startdoc1; ^$$ = '\\'; true; maindoc; {} {}
startdoc1; $$ = '\n'; true; StartDoc; {} {}
startdoc1; ^true; true; startdoc2; { "\\subsection[" } {}

startdoc2; $$ = '\n'; true; StartLine; { ']'; } {}
startdoc2; $$ = '*'; $2 = '*'; startdoc2_5; {} {}
startdoc2_5; $$ = '/'; true; to_eol; { ']'; } {}
startdoc2_5; ^true; true; startdoc2; { '*'; } {}
startdoc2; true; true; startdoc2; { $$; } {}

StartLine; $$ = $2; true; startline1; {} {}
StartLine; ^true; $2 != '*'; DocStrip; { "\\flush " } {}
StartLine; ^true; true; startline2; {} {}

startline1; $$ = $2; true; startline1; {} {}
startline1; ^true; true; startline2; {} {}

startline2; ($$ = ' ') || ($$ = '\t'); true; startline2; {} {}
startline2; $$ = '\n'; true; StartLine; { "\\par " } {}
startline2; $$ = '/'; $2 = '*'; to_eol; { "\\flush " } {}
startline2; ^true; true; maindoc; {} {}

maindoc; $$ = '\n'; true; StartLine; { ' ';} {}
maindoc; $$ = '\\'; true; after_slash; {} {}
maindoc; $$ = ' ' || $$ = '\t'; true; after_space; { ' '; }{}
maindoc; $$ = '*' && $2 = '*'; true; after_asterisk; {}{}
maindoc; true; true; maindoc; { $$; }{}

after_space; $$ = ' ' || $$ = '\t'; true; after_space; {}{}
after_space; $$ = '\n'; true; StartLine; {}{}
after_space; ^$$; true; maindoc; {}{}

after_slash; $$ = '\n'; true; skip_spaces; {}{}
after_slash; true; true; maindoc; { '\\'; $$; }{}
skip_spaces; $$ = ' ' || $$ = '\t'; true; skip_spaces; {}{}
skip_spaces; ^true; true; maindoc; {}{}

after_asterisk; $$ = '/'; true; to_eol; { "\\flush " }{}
after_asterisk; ^true; true; maindoc; { '*'; }{}

InError; true; true; InError; {}{}

.define StreamSplit 1000;
.define maybe_cmd 1004;
.define cmd_start 1001;
.define in_cmd 1003;
.define cmd_end 1002;
.define spaces_after_cmd 1005;
.final cmd_start;
.final cmd_end;


StreamSplit; $$ = '\\'; true; cmd_start; {}{}
StreamSplit; true; true; StreamSplit; { $$; }{}
cmd_start; $$@[=ID1] >= 0; true; in_cmd; { $$; }{}
cmd_start; true; true; cmd_end; { $$; }{}
in_cmd; $$ = ' '; true; spaces_after_cmd; {}{}
in_cmd; $$@[=ID] >= 0; true; in_cmd; { $$; }{}
in_cmd; ^true; true; cmd_end; {}{}
spaces_after_cmd; $$ = ' '; true; spaces_after_cmd; {}{}
spaces_after_cmd; ^true; true; cmd_end; {}{}

.define Arg 2000;
.define in_arg 2001;
.define arg_slash 2002;
.final Arg;

Arg; $$ = '['; true; in_arg; {} { 1; }
Arg; ^true; true; Arg; {}{}
in_arg; $$ = ']'; $1 = 1; Arg; {}{}
in_arg; $$ = ']'; true; in_arg; { $$; }{ $1 - 1; }
in_arg; $$ = '['; true; in_arg; { $$; }{ $1 + 1; }
in_arg; $$ = '\\'; true; arg_slash; { $$; } { $1; }
in_arg; true; true; in_arg; { $$; } { $1; }
arg_slash; true; true; in_arg; { $$; } { $1; }

.define Verbatim 3000;
.define end_verb 3001;
.define in_verb 3002;
.final end_verb;
Verbatim; true; true; in_verb; {}{ $$; }
in_verb; $$ = ']'; $1 = '['; end_verb; {}{}
in_verb; $$ = $1; true; end_verb; {}{}
in_verb; true; true; in_verb; { $$; }{ $1; }

.define ValidName 4000;
.define single_char 4001;
.define many_chars 4002;
.define invalid_name 4003;
.final ValidName;
ValidName; $$@[=ID1] >= 0; true; many_chars; {}{}
ValidName; true; true; single_char; {}{}
single_char; !$$; true; ValidName; {}{}
single_char; true; true; invalid_name; {}{}
many_chars; !$$; true; ValidName; {}{}
many_chars; $$@[=ID] >= 0; true; many_chars; {}{}
many_chars; true; true; invalid_name; {}{}

.define Cdecl 5000;
.define struct_like 5001;
.define func_like 5002;
.define func_like2 5003;
.define end_cdecl 5004;
.define end_cdecl1 5005;
.define cd_dl 5006;
.define skip_func_body 5010;
.final end_cdecl1;
Cdecl; $$ = '{'; true; struct_like; { $$; }{ 1; }
Cdecl; $$ = '('; true; func_like; { $$; } { 1; }
Cdecl; $$ = ';'; true; end_cdecl; { $$; }{}
Cdecl; true; true; Cdecl; { $$; }{}
struct_like; $$ = '{'; true; struct_like; { $$; } { $1 + 1; }
struct_like; $$ = '}'; $1 = 1; Cdecl; { $$; }{}
struct_like; $$ = '}'; true; struct_like; { $$; } { $1 - 1; }
struct_like; true; true; struct_like; { $$; }{ $1; }
func_like; $$ = '('; true; func_like; { $$; } { $1 + 1; }
func_like; $$ = ')'; $1 = 1; func_like2; { $$; } { }
func_like; $$ = ')'; true; func_like2; { $$; } { $1 - 1; }
func_like; true; true; func_like; { $$; } { $1; }
func_like2; $$ = '{'; true; skip_func_body; { ';'; }{ 1;}
func_like2; $$ = ';'; true; end_cdecl; { ';'; }{}
func_like2; $$ = '('; true; func_like; { $$; } { 1; }
func_like2; $$ = '\n'; true; cd_dl; {}{}
func_like2; true; true; func_like2; { $$; }{}
cd_dl; ^$$ = '{'; true; func_like2; {}{}
cd_dl; ^true; true; func_like2; {}{}
end_cdecl; $$ = '\n'; true; end_cdecl1; {}{}
end_cdecl; true; true; end_cdecl; {}{}
skip_func_body; $$ = '{'; true; skip_func_body; {} { $1 + 1; }
skip_func_body; $$ = '}'; $1 = 1; end_cdecl; {}{}
skip_func_body; $$ = '}'; true; end_cdecl; {} { $1 - 1; }
skip_func_body; true; true; skip_func_body; {} { $1; }

.define Cstat 5100;
.define cst_brace 5103;
.define end_stat 5101;
.define end_stat1 5102;
.final end_stat;
Cstat; $$ = '('; true; Cstat; { ($2 + 1):2, $$; }{}
Cstat; $$ = ')'; true; Cstat; { ($2 - 1):2, $$; }{}
Cstat; $$ = '{'; $2 = 0; cst_brace; { 1:2, $$; }{}
Cstat; $$ = ';'; $2 = 0; end_stat1; { $$; }{}
Cstat; true; true; Cstat; { $$; }{}
cst_brace; $$ = '{'; true; cst_brace; { ($2 + 1):2, $$;}{}
cst_brace; $$ = '}'; $2 = 1; end_stat1; { $$; }{} 
cst_brace; $$ = '}'; true; cst_brace; { ($2 - 1):2, $$; }{}
cst_brace; true; true; cst_brace; { $$; }{}
end_stat1; $$ = '\n'; true; end_stat; {}{}
end_stat1; true; true; end_stat1; {}{}

.define CID 5500;
.define in_id 5501;
.define cid_brace 5502;
.define cid_paren 5503;
.define cid_maybe_id 5504;
.define cid_after_id 5505;
.define cid_undo_id 5506;
.define cid_skip_bracket 5507;
.define cid_paren2 5508;
.define give_id 5509;
.define Reverse 5510;
.define give_rev 5511;
.define cid_void 5512;
.define cid_ignore 5513;
CID; $$@[=ID] >= 0; true; in_id; {} { 0; $$;}
CID; $$ = '('; true; cid_paren; {} { 1:2, $1; }
CID; $$ = '{'; true; cid_brace; {} { 1:2, $1; }
CID; $$ = '='; true; cid_ignore; {} { $1; }
CID; ^!$$; true; give_id; {} { $1; }
CID; true; true; CID; {}{ $1; }
cid_ignore; ^!$$; true; give_id; {} { $1; }
cid_ignore; true; true; cid_ignore; {}{ $1; }
in_id; $$@[=ID] >= 0; true; in_id; {} { $1; $$; }
in_id; ^true; true; CID; {} { $1; }
cid_brace; $$ = '{'; true; cid_brace; {} { ($2 + 1):2, $1; }
cid_brace; $$ = '}'; $2 = 1; CID; {} { 0:2, $1; }
cid_brace; $$ = '}'; true; cid_brace; {} { ($2 - 1):2, $1; }
cid_brace; true; true; cid_brace; {} { $1; }
cid_paren; $$ = '('; true; cid_paren; {} { ($2 + 1):2, $1; }
cid_paren; $$ = ')'; $2 = 1; CID; {} { 0:2, $1; }
cid_paren; $$ = ')'; true; cid_paren; {} { ($2 - 1):2, $1; }
cid_paren; $$ = 'v'; true; cid_maybe_id; {} { 1:4, $1; 0; $$; }
cid_paren; $$@[=ID] >= 0; true; cid_maybe_id; {}{ -1:4, $1; 0; $$; }
cid_paren; true; true; cid_paren; {} { $1; }
cid_maybe_id; $$ = 'o'; $4 = 1; cid_maybe_id; {} { 2:4, $1; $$; }
cid_maybe_id; $$ = 'i'; $4 = 2; cid_maybe_id; {} { 3:4, $1; $$; }
cid_maybe_id; $$ = 'd'; $4 = 3; cid_maybe_id; {} { 4:4, $1; $$; }
cid_maybe_id; $$@[=ID] >= 0; true; cid_maybe_id; {} { -1:4, $1; $$; }
cid_maybe_id; ^true; $4 = 4; cid_undo_id; {} { $1; }
cid_maybe_id; ^true; true; cid_after_id; {} { $1; }
cid_after_id; $$@[' ', '\t', '\n'] >= 0; true; cid_after_id; {}{ $1; }
cid_after_id; $$ = '['; true; cid_skip_bracket; {} { $1; }
cid_after_id; ^$$ = ')'; true; cid_paren2; {} { $1; }
cid_after_id; ^true; true; cid_undo_id; {} { $1; }
cid_undo_id; ^true; $1;  cid_undo_id; {} {}
cid_undo_id; ^true; true; cid_paren2; {} {}
cid_skip_bracket; $$ = ']'; true; cid_after_id; {}{ $1; }
cid_skip_bracket; true; true; cid_skip_bracket; {} { $1; }
cid_paren2; $$ = '('; true; cid_paren2; {} { ($2 + 1):2, $1; }
cid_paren2; $$ = ')'; $2 = 1; CID; {} { 0:2, $1; }
cid_paren2; $$ = ')'; true; cid_paren2; {} { ($2 - 1):2, $1; }
cid_paren2; true; true; cid_paren2; {}{ $1; }
give_id; ^true; $1; give_id; { $1; }{}
give_id; true; true; CID; { 0; }{}
Reverse; ^!$$; true; give_rev; { $1; }{}
Reverse; true; $#;  Reverse; {}{ $1; $$; }
Reverse; true; true; Reverse; {} { $$; }
give_rev; ^true; $#; give_rev; { $1; } {}
give_rev; true; true; give_rev; {}{}

.define SourceIndex 6000;
.define pre_index 6001;
.define in_index 6002;
.define IndexTable 6003;
.define in_table 6004;
.define index_slash 6005;
.final SourceIndex;
.final IndexTable;

SourceIndex; $$ = '\"'; true; in_index; {}{}
SourceIndex; true; true; pre_index; {}{}
pre_index; $$ = '\"'; true; in_index; {}{}
pre_index; true; true; pre_index; {}{}
in_index; $$ = '\"'; true; SourceIndex; {}{}
in_index; $$ = '\\'; true; index_slash; {}{}
in_index; true; true; in_index; { $$; }{}
index_slash; $$ = '\\'; true; in_index; { "\\\\" }{}
index_slash; $$ = '\"' || $$ = '\''; true; in_index; { $$; }{}
index_slash; true; true; in_index; { '\\'; $$; }{}

IndexTable; true; true; in_table; { $$; } { 0; }
in_table; $$ = '{'; true; in_table; { $$; } { $1 + 1; }
in_table; $$ = '}'; $1 < 1; IndexTable; { $$; }{}
in_table; $$ = '}'; true; true; { $$; } { $1 - 1; }
in_table; true; true; in_table; { $$; }{ $1; }

.define Toalpha 7000;
.define ToAlpha 7001;
.define Toroman 7002;
Toalpha; $$ > 0 && $$ <= 26; true; Toalpha; { ['a'-'z']($$ - 1); }{}
Toalpha; true; true; Toalpha; {}{}
ToAlpha; $$ > 0 && $$ <= 26; true; ToAlpha; { ['A'-'Z']($$ - 1); }{}
ToAlpha; true; true; ToAlpha; {}{}

.define Filter 8000;
Filter; $$@['0'-'9'] >= 0; true; Filter; { $$; }{}
Filter; true; true; Filter; { ' ';}{}
.end

main
{
	global 	mode = user 
			output-type = man
			global-options
			active-stream
			prev-stream
			active-line
			active-mode = user
			condition = true
			if-cnt
			false-if-cnt
			has-text
			divert-cmd
			par-cmd
			in-list
			relax
			last-anchor
			verbose quiet
			this-file
			outprefix
			format-file = docstrip.fmt
			doc-path = "./\:sysvar(env:location)scripts/docstrip\:sysvar(env:location)"
			doc-fmt
			cdecl-id
			strong-commands = list(flush if ifx ifkept ifdef ifuser iftext ifdev ifnum ifout 
									endif else onflush onpar divert par user dev any restore)
			embed-commands = list(begin insert if ifx ifdef iftext ifnum else eat endif 
								cdeclid thisfile bodyof symbol gensym
								add divide length alpha Alpha number unnumber lastanchor repeat);
	table macros ''
	getopt(list(type = : 
				t = =type 
				help
				mode = :
				m = =mode
				verbose
				v = =verbose
				quiet
				q = =quiet
				options = : 
				o = =options
				format = :
				f = =format
				docpath = :
				p = =docpath
				docfmt = :
				define = :
				D = =define
				debug
				dir = =directory
				directory = :
				outdir = :
				) 
			process-getopt)
	if not(instream())
		defstream stdin '' basic;
		instream stdin
	endif
	if eq(outstream() stdout)
		load-type($output-type)
		outstream defstream(* '' :$output-type true) 
	endif
	if $quiet
		ioctl outstream() quiet;
	endif
	send-options($global-options)
	defautoseq rl linefilter
	defautoseq parse docstrip:StreamSplit;
	defautoseq arg docstrip:Arg;
	defautoseq verbatim docstrip:Verbatim
	defautoseq checkname docstrip:ValidName
	defautoseq docstrip docstrip;
	defautoseq cdecl docstrip:Cdecl;
	defautoseq cdeclid docstrip:CID|docstrip:Reverse
	defautoseq cstmt docstrip:Cstat;
	defautoseq indexitem docstrip:SourceIndex;
	defautoseq indextable docstrip:IndexTable
	defautoseq toalpha intconv:FromInt|docstrip:Toalpha;
	defautoseq toAlpha intconv:FromInt|docstrip:ToAlpha;
	defautoseq tonumber docstrip:Filter|intconv:FromInt|unicode:Tohex
	table keeps ''
	def line text cmd;
	set active-stream = instream()
	process-line('\\any\\relax\\include['$format-file']\\restore')
	while;
		set line = read(reset(docstrip))
		if and(not($line) iseof())
			clear relax
			if $verbose echo '[eof]' endif
			pop active-stream
			pop this-file
			if not($active-stream) break endif
			instream $active-stream
			set line = $active-line
			pop active-line
		endif
		if eq(statename(docstrip '') InError) error "syntax error" endif
		process-line($line)
	loop;
}

#### \section[SYNOPSIS]
### \sample[docstrip -\|-mode=[user|dev[eloper]] -\|-type=[man|plain|html] -\|-options=global-opts
### -\|-verbose -\|-quiet -\|-format=fmt-file-name -\|-docpath=add-doc-path -\|-docfmt=self-doc-file
### -\|-define=macro-name[=value] -\|-dir[ectory]=cur-dir -\|-outdir=output-dir -\-debug
### [infilename [outfilename]]]
### \list
### \item[mode] chooses developer or user documentation (default \em[user])
### \item[type] chooses output format (default \em[man])
### \item[options] global options to pass to a DocStrip backend
#### \item[verbose] print \\trace output
### \item[quiet] do not issue error and warning messages
### \item[format] use format file \em[fmt-file-name] (default \em[docstrip.fmt])
### \item[docpath] add \em[add-doc-path] to a path where DocStrip files are searched for
### Multiple options are cumulative. The default path is the current directory, then
#### \sample[scripts/docstrip] within Tensile location, then the top of Tensile location.
### \item[define] predefine a macro \em[macro-name] with \var[value] as its body
### \item[directory] changes directory to \em[new-dir]
### \item[outdir] uses \em[output-dir] as a prefix for output filenames
### (note that the prefix is added even if a filename has a directory part)
### \item[docfmt] specifies a file where documentation on defined macros should go
### \item[debug] turns of issuing debugging info (for developers only)
### \endlist
### If no input filename is given, \em[stdin] is assumed; if no output filename is
### given, \em[stdout] is assumed
process-getopt(type option arg extra)
{
	switch $type
		when eq($^ option)
			switch $option
				when eq($^ options)
					set global-options = $arg
					breaksw
				when eq($^ mode)
					if and(neq($arg user) not(isprefix(developer $arg)))
						error "invalid mode: $arg"
					endif
					set mode = cond(eq($arg user) user dev)
					breaksw
				when eq($^ verbose)
					set verbose = true
					breaksw
				when eq($^ quiet)
					set quiet = true
					set verbose = ''
					breaksw
				when eq($^ type)
					set output-type = $arg
					breaksw
				when eq($^ format)
					set format-file = $arg
					breaksw
				when eq($^ docpath)
					set doc-path = $doc-path\:$arg
					breaksw
				when eq($^ docfmt)
					set doc-fmt = defstream(* file://$arg basic true)
					breaksw
				when eq($^ define)
					def n v;
					set v = split($arg '=' '' n)
					if not(check(macros $n))
						insert macros $n 0:$v
					else
						put macros $n 0:$v
					endif
					breaksw
				when eq($^ debug)
					trace *
					breaksw
				when eq($^ directory)
					cd $arg
					breaksw
				when eq($^ outdir)
					set outprefix = $arg
					if not(issuffix($outprefix /))
						set outprefix = $outprefix/
					endif
					breaksw					
				when eq($^ help)
					echo 
'Usage: docstrip [--type=out-type] [--mode={user,developer}] [--options=opts] [in-file] [out-file]'
					exit()
			endsw
			breaksw
		when eq($^ name)
			if not(instream())
				set this-file = $option
				set option = searchpath($option $doc-path '')
				if not($option) error "file $this-file not found" endif
				defstream stdin file://$option basic;
				instream stdin
			elsif(neq(outstream() stdout))
				error "too many filenames"
			else
				load-type($output-type)
				outstream defstream(* $outprefix$option :$output-type true)
			endif	
			breaksw
		when true
			error "unknown option --$option"
	endsw
}

#### \part\section[GENERAL STRUCTURE]
#### DocStrip input is embedded into a source file inside comments. 
#### It support both C and C++ style comments as well as traditional shell hash comments.
#### DocStrip block starts with \sample[///], \sample[/***] or \sample[###].
#### The block ends at the first non-comment line or at enclosing \sample[*/].
#### Comment-introducing characters, asterisks and spaces at line beginnings are 
#### swallowed. Non-empty lines are joined together. Line continuations (aka \sample[\\])
#### are recognized.
####
#### Essentially, a DocStrip block is a sequence of text characters and
#### macros introduced with a backslash. Macros may have arguments which are
#### enclosed into brackets. A macro name is either a single character or a
#### sequence of alphanumeric characters and underscores starting with a letter.
#### If a backslashed character has no macro definition, it is treated as an ordinary
#### After alphanumeric macro names spaces are swallowed. Whether they are kept after
#### arguments is macro-dependant.
####
#### If a first line of a block doesn't start with a macro, it is treated as a
#### subsection header.
####
#### Empty lines separate paragraphs.
####
#### There is a set of DocStrip primitives. Other common macros are defined in a
#### \em[format file]. Some primitives expand their arguments using a limited
#### subset of primitives. User-defined macros are expanded with arguments textually
#### substituted, and the result of expansion is then re-examined.
#### \section[SPECIAL CHARACTERS]
#### Some characters and character sequences are treated specially by DocStrip backends
#### They include:
#### \item[~~] non-breaking space (to get a real tilde, type it twice)
#### \item[-\|-] en dash --
#### \item[-\|-\|-] em dash --- 
#### \item[(\|c)] copyright sign (c)
#### \item[-\|>] single right arrow ->
###  \item[-\|-\|>] same 
#### \item[<\|-] single left arrow <-
#### \item[<\|-\|-] same
#### \item[=\|>] greater or equal =>
#### \item[>\|=] same
#### \item[<\|=] less or equal <=
#### \item[==\|>] double right arrow ==>
#### \item[<\|==] double left arrow <==
#### \item[<\|-\|>] single bidi arrow <->
#### \item[<\|=\|>] double bidi arrow <=>
#### \item[!\|=] not equal !=
#### \item[+\|-] plus-or-minus +-
#### \item[DEL] (ASCII 127) a special non-joiner character which is used to break the above
#### ligatures
process-line(line)
{
	def text cmd;
	push(active-line = $line)
	while $active-line
		set active-line = ausplit($active-line reset(parse) <none> text cmd)
		if $text set has-text = true endif
		if and(not($relax) eq($active-mode $mode) $condition) write $text endif
		if $cmd
			if check(macros $cmd)
				expand-macro($cmd)
				continue
			endif
			if and(eq(int length($cmd) 1) not(match($cmd '[a-zA-Z]')))
				if and(not($relax) eq($active-mode $mode) $condition) write char(0x7F)$cmd endif
			else
				if or(and($condition eq($mode $active-mode)) check($strong-commands $cmd))
					call(docstrip_$cmd)
				endif
			endif
		endif
	loop
	pop active-line
}

process-macros(line)
{
	def text cmd output;
	push(active-line = $line)
	push(condition = true)
	push(has-text = '')
	while $active-line
		set active-line = ausplit($active-line reset(parse) <none> text cmd)
		if $condition set output = $output$text if $output set has-text = true endif endif
		if $cmd
			if check(macros $cmd)
				expand-macro($cmd)
			elsif eq($cmd protect)
				set active-line = ausplit($active-line reset(arg) <none> text)
				set output = $output$text
				set has-text = true
			elsif check($embed-commands $cmd)
				if or($condition check($strong-commands $cmd))
					call(docstrip_$cmd) 
				endif
			else
				set output = $output\\$cmd
				set has-text = true
			endif
		endif
	loop
	pop has-text
	pop condition
	pop active-line
	return($output)
}


send-options(opts)
{
	def opt;
	while $opts
		set opts = split($opts ; '' opt)
		set opt = trim($opt)
		if $opt
			ioctl outstream() option=$opt
		endif
	loop
}

expand-macro(cmd)
{
	def macro = get(macros $cmd) argcnt i arg;
	set macro = split($macro : '' argcnt);
	local array(args)
	reset arg
	if neq(int $argcnt 0)
		foreach i local(iota($argcnt))
			set active-line = ausplit($active-line arg <none> arg)
			insert args $i $arg
		loop
	endif
	set active-line = expand($macro args '#()##')$active-line
}
#### \part\section[PRIMITIVES]
#### Macro arguments are given in the line below its name.
#### Each argument name is enclosed in brackets. Arguments which should not be empty
#### are marked with an exclamation mark. Expandable arguments are marked with an asterisk.
#### If a missing/empty argument is considered to have some default value, it is given 
#### after \em[=] sign. 
#### 
#### If a primitive itself is expanded in arguments, it is indicated with a plus after an
#### argument list. If a primitive swallows spaces even after arguments, it is indicated with
#### a \em[@].
#### \note DocStrip have some other primitives besides those listed below,
#### but they are intended only for format-file developers and thence are undocumented
#### here

docstrip_emit()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg)
	if not($arg) error "missing arg for \\emit" endif
	writetag $arg;
}

#### href
#### [description*][location*]
####
#### Emits a hyper-reference whose user-visible part is \var[description] and
### which points to \var[location]. If a \var[location] is empty, it is guessed from
#### the \var[description]. In particular, man page references in the form
### \sample[id(section)] are recognized. 
### A \var[location] may be either an URL or a anchor reference starting with \em[#].
#### In the former case, if a \var[description] is empty, it is considered equal to
#### \var[location]. Otherwise, the result is backend-dependant.
docstrip_href()
{
	def arg location;
	set active-line = ausplit($active-line reset(arg) <none> arg location)
	writetag HREF VALUE process-macros($arg) LOCATION process-macros($location);
}

### anchor
### [name*]
###
### Sets an anchor to be refenced by \em[href] at the current point.
### If \var[name] is empty, a unique name is generated.
docstrip_anchor()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg)
	writetag ANCHOR NAME set(last-anchor cond($arg process-macros($arg) 'a_'substr(gensym() 1)));
}

#### lastanchor
#### +
####
#### Expands to the most recently defined anchor name (useful with generated names)
docstrip_lastanchor()
{
	set active-line = $last-anchor$active-line;
}

#### gensym 
#### +
####
#### Expands to a unique name suitable e.~g. for anchor names 
docstrip_gensym()
{
	set active-line = substr(gensym() 1)$active-line;
}

#### verbatim
####
#### This primitive takes the next input character as a delimiter and then
#### treats all the following characters up to the next delimiter as ordinary.
#### Special characters are not recognized within \em[verbatim]
docstrip_verbatim()
{
	def verb;
	set active-line = ausplit($active-line reset(verbatim) <none> verb)
	writetag PRE INLINE true;
	write $verb
	writetag /PRE
}

#### begin
#### +
####
#### Sets a mark at the current point (not an anchor!)
docstrip_begin()
{
	clear has-text
}

docstrip_relax()
{
	set relax = true
}

#### define
#### [name!][body][argcnt][doc-string] @
####
#### Defines a macro \var[name] with \var[body]. The macro will accept
#### at most \var[argcnt] arguments. A body may contain \em[argument references]
#### which are given as \sample[#number] or \sample[#(number)]. Literal hash
#### should be doubled.
####
#### A \var[name] should not be defined.
####
#### If \var[doc-string] is not empty and a self-documentation file is defined,
#### \var[doc-string] is written to that file prepended by a macro name
#### as a subsection.
docstrip_define()
{
	def name body argcnt docstr;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> name body argcnt docstr)
	if not($name) error "missing macro name in \\define" endif
	if check(macros $name) error "macro $name already defined" endif
	translatez(reset(checkname) $name)
	if not(isfinal(checkname)) error "invalid name $name" endif
	if not(isint($argcnt)) error "invalid argument count $argcnt for $name" endif
	insert macros $name cond($argcnt $argcnt 0):$body
	if and($docstr $doc-fmt)
		writeto $doc-fmt "\n#\#\# $name\n#\#\# $docstr\n"
	endif
}

#### redefine
#### [name][body] @
####
#### Redefines an alredy-defined macro to be \var[body].
#### Argument count remains the same.
docstrip_redefine()
{
	def name body argcnt;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> name body)
	if not(check(macros $name)) error "macro $name is not defined" endif
	split(get(macros $name) : '' argcnt) 
	put macros $name $(argcnt)\:$body
}

#### undefine
#### [name]
####
#### Undefines a macro \var[name]. If there is no such macro, does nothing.
docstrip_undefine()
{
	def name;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> name)
	if check(macros $name)
		remove macros $name
	endif
}

#### let
#### [name][body*] @
####
#### Like \em[redefine], but expands \var[body]
docstrip_let()
{
	def name body argcnt;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> name body)
	if not(check(macros $name)) error "macro $name is not defined" endif
	split(get(macros $name) : '' argcnt) 
	put macros $name $(argcnt)\:process-macros($body)
}

#### bodyof
#### [name] +
####
#### Expands to the body of a macro \var[name] or to an empty string if it is not defined
#### No argument substitutions occur.
docstrip_bodyof()
{
	def name;
	set active-line = ausplit($active-line reset(arg) <none> name);
	if check(macros $name)
		set active-line = split(get(macros $name) : '' '')$active-line
	endif
}

#### repeat
#### [count*][body] +
#### 
#### Expands to \var[body] repeat \var[count] times.
docstrip_repeat()
{
	def count body;
	set active-line = ausplit($active-line reset(arg) <none> count body)
	set active-line = repeat($body process-macros($count))$active-line
}

#### par
####
#### Starts a new paragraph
docstrip_par()
{
	if not($condition) return endif
	set active-line = $par-cmd$active-line
	if eq($mode $active-mode) writetag P; endif
}


#### indent
#### [label*] @
####
#### Starts an indent block of text marked with \em[label] (if not empty).
#### Sequential calls of \em[indent] are \strong[not] cumulative
docstrip_indent()
{
	def label;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> label)
	writetag INDENT LABEL process-macros($label);
}

#### noindent
####
#### Cancels indenting caused by \em[indent]
docstrip_noindent()
{
	writetag NOINDENT;
}

#### list
####
#### Starts a list
docstrip_list()
{
	push(in-list = true);
	writetag LIST;
}

#### item
#### [label*]
####
#### Starts a new list item (and a new list if there isn't one) possibly marking it
#### with \var[label]. If there is no label, the marker is backend-specific, but
#### usually a bullet will be used.
docstrip_item()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg)
	if not($in-list) writetag LIST; set in-list = true endif
	if $arg
		set active-line = '\\eat '$active-line
		writetag ITEM LABEL process-macros($arg);
	else
		writetag ITEM;
	endif
}

#### endlist
#### 
#### Ends a list. A list is also ended by the end of the block.
#### Some backends may end a list accepting \em[indent] or sectionig commands.
docstrip_endlist()
{
	if not($in-list) error "\\endlist without \\item" endif
	pop in-list
	writetag /LIST;
}

#### endlists
####
#### Ends all the active lists.
docstrip_endlists()
{
	while $in-list
		pop in-list
		writetag /LIST;
	loop
}

#### options
#### [opts] @
####
#### Sends \em[opts] to the current backend. Options are a set of semicolon-separated
#### pairs \sample[keyword=value]
docstrip_options()
{
	def type options;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> type options)
	if or(not($type) eq($type $output-type))
		send-options($options)
	endif
}

#### divert
#### [cmds!] @
####
#### Starts a diversion, so that all the following output is kept at a special location.
#### \var[cmds] are to be executed after the diversion end.
#### Diversions cannot nest.
docstrip_divert()
{
	if $divert-cmd error "diversions cannot nest: $divert-cmd" endif
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> divert-cmd)	
	if not($divert-cmd) error "missing diversion flush cmd" endif
	if $condition
		ioctl outstream() divert
	else
		clear divert-cmd
	endif
}

#### onflush
#### [cmds]
#### 
#### Arranges \var[cmds] to be executed at the next flush (see \em[flush])
docstrip_onflush()
{
	def cmd;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> cmd)
	if $condition set divert-cmd = $divert-cmd$cmd endif
}

#### onpar
#### [cmds]
####
#### Arranges \var[cmds] to be executed after the end of each paragraph
docstrip_onpar()
{
	def cmd;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> cmd)
	if $condition set par-cmd = $par-cmd$cmd endif
}

#### flush
####
#### Ends the current diversion if their is one, and executes \em[onflush]ed commands
#### \note An implicit \em[flush] occurs at the end of a block
docstrip_flush()
{
	if not($condition) return endif
	if neq($mode $active-mode) set active-line = $divert-cmd$active-line return endif
	docstrip_endlists()
	if $divert-cmd
		ioctl outstream() nodivert	
		set active-line = $divert-cmd$active-line
		clear divert-cmd
	endif
}

#### insdiv
####
#### Issues the content of a previously defined diversion
docstrip_insdiv()
{
	ioctl outstream() insdiv
}

incr-if()
{
	incr if-cnt 1
	if not($condition) incr false-if-cnt 1 endif
}

#### if
#### [val1][val2] @+
#### 
#### Skips all the input until a matching \em[else] or \em[endif]
### if \var[val1] != \var[val2]
docstrip_if()
{
	def val1 val2;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> val1 val2);
	if $condition
		set condition = eq($val1 $val2)
	endif
	incr-if()
}

#### ifx 
#### [val1*][val2*] @+
####
#### Same as \em[if], but arguments are expanded
docstrip_ifx()
{
	def val1 val2;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> val1 val2);
	if $condition
		set condition = eq(process-macros($val1) process-macros($val2))
	endif
	incr-if()
}

#### ifnum
#### [op][num1][num2] @+
####
#### Performs a numerical comparison on expanded strings \var[num1] and \var[num2].
#### \var[op] may be \em[<], \em[>] or \em[=].
docstrip_ifnum()
{
	def op val1 val2;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> val1 op val2);
	if $condition
		switch $op
			when eq($^ =)
				set condition = eq(int process-macros($val1) process-macros($val2))
				breaksw
			when eq($^ <)
				set condition = less(int process-macros($val1) process-macros($val2))
				breaksw
			when eq($^ >)
				set condition = greater(int process-macros($val1) process-macros($val2))
				breaksw
			when true
				error "invalid op $op"
		endsw
	endif
	incr-if()
}

#### iftext
#### +
####
#### True if there was a text after the last \em[begin] or the start of the argument
docstrip_iftext()
{
	if $condition
		set condition = $has-text
	endif
	incr-if()
}

#### ifdef
#### [name] @+
####
#### True if a macro \var[name] is defined
docstrip_ifdef()
{
	def name;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> name);
	if $condition
		set condition = check(macros $name)
	endif
	incr-if()
}

#### ifkept
#### [kind][item*] @+
####
#### True if there is a stored \var[item] in a table \var[kind].
docstrip_ifkept()
{
	def kind item;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> kind item);
	if $condition
		set kind = $kind\:process-macros($item)
		def handle;
		set item = sto.enum(keeps handle $kind)
		set condition = isprefix($item $kind)
		if $handle sto.break(keeps handle) endif
	endif
	incr-if()
}

#### ifout
#### [type] @+
####
#### True if the current output type is \var[type]
docstrip_ifout()
{
	def out;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> out)
	if $condition
		set condition = eq($output-type $out)
	endif
	incr-if()
}

#### ifuser
#### +
####
#### True if in user doc mode
docstrip_ifuser()
{
	if $condition
		set condition = eq($mode user)
	endif
	incr-if()
}

#### ifdev
#### +
#### 
#### True if in developer doc mode
docstrip_ifdev()
{
	if $condition
		set condition = eq($mode dev)
	endif
	incr-if()
}

#### else
#### +
####
#### See \em[if]
docstrip_else()
{
	if less(int $if-cnt 1) error "\\else without \\if..." endif
	if $condition
		set condition = ''
		set false-if-cnt = 1
	elsif eq(int $false-if-cnt 1)
		set condition = true
		decr false-if-cnt 1
	endif
}

#### endif
#### +
####
#### See \em[if]
docstrip_endif
{
	decr if-cnt 1
	if not($condition) decr false-if-cnt 1 endif
	if less(int $if-cnt 0) error "\\endif without \\if..." endif
	if eq(int 0 $false-if-cnt) set condition = true endif
}

#### include
#### [name] @
####
#### Includes a file \var[name] which is searched in the document path if
#### doesn't start with `/', `./' or `../'
docstrip_include()
{
	def filename;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> filename);
	if not($filename) error "filename is missing" endif
	push(this-file = $filename)
	set filename = searchpath($filename $doc-path '')
	if not($filename) error "file $this-file not found" endif
	if $verbose echo \[$filename\] endif
	push(active-stream = defstream(* $filename basic))
	push(active-line = '')
	instream($active-stream)
}

#### output
#### [type][name*][oprions*]
####
#### Uses a file \var[name] as a new output file of the \var[type],
#### sending it \var[options]. The current output file is remembered unless
#### the first char of \var[name] is `='
####
#### If \var[type] is empty, the current type is assumed.
#### If \var[name] is empty, the output is switched to the \em[previous] output file;
#### if there is no such file, nothing happens.
#### If \var[options] start with `;', options passed on command line are prepended
docstrip_output
{
	def type name options;
	ioctl outstream() flush
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> type name options)
	set name = process-macros($name)
	set options = process-macros($options)
	if not($type) set type = $output-type endif
	if $name
		if isprefix($name =)
			rmstream outstream()
			set name = substr($name 1)
		else
			push(prev-stream = outstream())
		endif
		load-type($type)
		if $verbose echo "\[output to $name\]" endif
		outstream defstream(* $outprefix$name :$type true)
		if $quiet
			ioctl outstream() quiet
		endif
	else
		if $prev-stream
			if $verbose echo "\[output to previous]" endif
			rmstream outstream()
			outstream $prev-stream
			pop prev-stream
		endif
	endif
	if isprefix($options ;)
		set options = $global-options$options
	endif
	send-options($options)
}

#### echo
#### [msg] @
####
#### Issues a \var[msg] to \em[stderr] if not in quite mode
docstrip_echo
{
	def msg;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> msg)
	if not($quiet)
		echo process-macros($msg)
	endif
}

#### trace
#### [msg*] @
####
#### Issues a \var[msg] to \em[stderr] if in verbose mode
docstrip_trace
{
	def msg;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> msg)
	if $verbose echo \[process-macros($msg)\] endif
}

#### cdecl
####
#### Extracts a C-like declaration from source following the end of the block and
#### outputs it
docstrip_cdecl()
{
	def tmp;
	writetag PRE;
	write set(tmp read(reset(cdecl)));
	set cdecl-id = translatez(reset(cdeclid) $tmp)
	writetag /PRE;
}

### cstmt
###
### Like the previous, but operates on C statements
docstrip_cstmt()
{
	writetag PRE;
	write read(reset(cstmt))
	writetag /PRE;
}

#### cdeclid
#### +
####
#### Expands to the identifier caught by the last \em[cdecl]
docstrip_cdeclid()
{
	set active-line = $cdecl-id$active-line
}

#### thisfile
#### +
####
#### Expands to the name of the current input file
docstrip_thisfile()
{
	set active-line = $this-file$active-line
}

#### source
#### [cnt] @
####
#### Extracts up to \var[cnt] lines of source following the end of the block and
#### outputs them.
#### A line containing \em[\\end] terminates extraction.
#### If \var[cnt] is empty, an infinite maximal number is assumed.
#### \note The standard format file redefines this primitive so that
#### it enforces developer mode
docstrip_source()
{
	def line count;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> count);
	writetag PRE;
	while;
		set line = read(rl)
		if or(iseof() index($line \\end)) break endif
		write $line\n
		if $count decr count 1 endif
	loop(or(not($count) neq(int $count 0)));
	writetag /PRE;
}

#### sourceindex
#### [kind] @
####
#### The source code that follows the block is treated as an array of string literals.
#### Those are extracted and stored in a table \var[kind]
#### \bug No escape sequences in literals are recognized

docstrip_sourceindex()
{
	def index = read(indextable) item kind;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> kind)
	while $index;
		set index = ausplit($index indexitem <none> item)
		if and($item not(check(keeps $kind\:$item)))
			insert keeps $kind\:$item ''
		endif
	loop
}

#### keep
#### [table][key*][property]
####
#### Stores a \var[key] associated with a \var[property] in a \var[table].
#### If a \var[key] has been already stored, the property is overwritten.
docstrip_keep()
{
	def name value property;
	set active-line = ausplit($active-line reset(arg) <none> name value property)
	set value = process-macros($value)
	if not(check(keeps $name\:$value))
		insert keeps $name\:$value $property
	elsif $property
		put keeps $name\:$value $property
	endif
}

#### set
#### [table][key*][property*]
####
#### Like the previous, but expands its second argument
docstrip_set()
{
	def name value property;
	set active-line = ausplit($active-line reset(arg) <none> name value property)
	set value = process-macros($value)
	set property = process-macros($property)
	if not(check(keeps $name\:$value))
		insert keeps $name\:$value $property
	else
		put keeps $name\:$value $property
	endif
}

#### insertall
#### [table][cmds][sep-cmds][prefix*] @+
####
#### Expands to all the records from a \var[table] whose keys start with a \var[prefix].
#### \var[cmds] are executed for each such record, substituting a given key for
#### \code[%key;] and its property for \code[%prop;]. Literal % should be typed twice
#### inside \var[cmds]. \var[sep-cmds] are inserted between records (there is no
#### substitutions here)
docstrip_insertall()
{
	def name cmds sepcmd handle key prefix;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> name cmds sepcmd prefix)
	set prefix = $name\:process-macros($prefix)
	set key = sto.enum(keeps handle $prefix)
	if not(isprefix($key $prefix)) return endif
	def params = local(list(key prop));
	put $params key split($key : '' '')
	put $params prop get(keeps $key)
	process-line(expand($cmds $params '%();%'))
	while;
		set key = sto.enum(keeps handle)
		if not(isprefix($key $prefix))
			if $handle 
				sto.break(keeps handle) 
			endif 
			break 
		endif
		put $params key split($key : '' '')
		put $params prop get(keeps $key)
		process-line($sepcmd)
		process-line(expand($cmds $params '%();%'))
	loop;
}

#### insert
#### [table][key*] +
####
#### Expands to a property associated with a \var[key] in a \var[table]; or
### to an empty string if no such key is stored.
docstrip_insert()
{
	def name id;
	set active-line = ausplit($active-line reset(arg) <none> name id);
	set id = process-macros($id)
	if check(keeps $name\:$id)
		set active-line = get(keeps $name\:$id)$active-line
	endif
}

#### zap
#### [table] @
###
#### Clears a \var[table]. If \var[table] is empty, clears all tables
docstrip_zap()
{
	def idx handle key;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> idx)
	if not($idx)
		release storage keeps
		table keeps ''
		return
	endif
	while;
		set key = sto.enum(keeps handle $idx\:)
		if not($handle) break endif
		mark.remove(keeps $key)
	loop;
	do.remove(keeps) 
}

#### add
#### [val1*][val2*] +
####
#### Expands to the sum of \var[val1] and \var[val2]
docstrip_add()
{
	def val1 val2;
	set active-line = ausplit($active-line reset(arg) <none> val1 val2);
	set active-line = add(process-macros($val1) process-macros($val2))$active-line
}

#### divide
#### [val1*][val2*] +
####
#### Expands to the integral quotient of \var[val1] and \var[val2]
docstrip_divide()
{
	def val1 val2;
	set active-line = ausplit($active-line reset(arg) <none> val1 val2);
	set active-line = div(process-macros($val1) process-macros($val2))$active-line
}

#### length
#### [val*] +
####
#### Expands to the length of expanded \var[val]
docstrip_length()
{
	def val1;
	set active-line = ausplit($active-line reset(arg) <none> val1);
	set active-line = length(process-macros($val1))$active-line
}

#### alpha
#### [num*] +
####
#### Expands to \var[num]th Latin lower-case letter
docstrip_alpha()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg);
	set active-line = translatez(reset(toalpha) process-macros($arg))$active-line
}

#### Alpha
#### [num*] +
####
#### Expands to \var[num]th Latin upper-case letter
docstrip_Alpha()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg);
	set active-line = translatez(reset(toAlpha) process-macros($arg))$active-line
}

#### number
#### [val*] +
####
#### Expands to an opaque number representation suitable for correct sorting
docstrip_number()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg);
	set active-line = translatez(reset(tonumber) process-macros($arg))$active-line
}

#### unnumber
#### [num*] +
####
#### Converts that opaque representation back to arabic digits
docstrip_unnumber()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg);
	set active-line = dec('0x'process-macros($arg))$active-line
}

docstrip_NSLversion()
{
	write version();
	set has-text = true
}

#### date
####
#### Issues the current date
docstrip_date()
{
..ifdef timestamp
	write ascii-time(timestamp())
	set has-text = true
..endif	
}

#### symbol
#### [sym] +
####
#### Issues a character with ASCII code \var[sym]
docstrip_symbol()
{
	def ch;
	set active-line = ausplit($active-line reset(arg) <none> ch);
	set ch = char($ch);
	if neq($ch \\) set active-line = $ch$active-line else set active-line = \\\\$active-line endif
}

### eat
###
# a null primitive used to eat unnecessary spaces
docstrip_eat()
{
}

#### builtin
#### [cmd]
####
#### Executes \var[cmd] as a primitive even if it is shadowed by a macro
docstrip_builtin()
{
	def cmd;
	set active-line = ausplit($active-line reset(arg) <none> cmd)
	call(docstrip_$cmd)
}

##### protect
##### [arg] +
####
#### Prevents its argument from expansion 
docstrip_protect()
{
	def arg;
	set active-line = ausplit($active-line reset(arg) <none> arg)
	set active-line = $arg$active-line
}

##### user
#####
##### Marks the following doc as user's
docstrip_user()
{
	if $condition push(active-mode = user) endif
}

##### dev
#####
##### Marks the following doc as developer's
docstrip_dev()
{
	if $condition push(active-mode = dev) endif
}

##### any
####
#### Marks the following doc as ambivalent
docstrip_any()
{
	if $condition push(active-mode = $mode) endif
}

##### restore
#####
##### Restores the user/developer mode before the last \em[user], \em[dev] or \em[any]
docstrip_restore()
{
	if $condition
		pop active-mode
		if not($active-mode) set active-mode = user endif
	endif
}

#### fmtdoc
#### [self-doc] @
####
#### Writes \var[self-doc] into a self-documentation file if it is defined;
#### otherwise does nothing
docstrip_fmtdoc()
{
	def arg;
	set active-line = '\\eat 'ausplit($active-line reset(arg) <none> arg)
	if $doc-fmt
		writeto $doc-fmt "#\#\# $arg\n"
	endif
}

load-type(type)
{
	once(table(types ''))
	if not(check(types $type))
		subload :$type docstrip_$type
		insert types $type ''
	endif
}

#### \part\section[STANDARD FORMAT MACROS]
#### \include[docstrip.fmt.doc]
#### \trailer[docstrip]
