#
# 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#

.module <readline>
main
{
	def tbl_name;
	global handler cur_item no-confirm;
	table toedit set(tbl_name get(ARGV 1))
	def cmd rest;
	set cur_item = sto.enum(toedit handler)
	while;
		set cmd = readline($tbl_name\:$cur_item>)
		set rest = split($cmd ' ' '' cmd)
		if not(isproc(cmd_$cmd))
			echo "Invalid command: $cmd"
		else
			call cmd_$cmd $rest;
		endif
	loop
}

cmd_quit
{
	exit
}

cmd_exit
{
	exit
}

cmd_list(arg)
{
	if not($arg) set arg = * endif
	def iter;
	foreach iter toedit
		if match($arg $iter)
			echo "$iter\: get(toedit $iter)"
		endif
	loop
}

cmd_current
{
	echo get(toedit $cur_item)
}

cmd_count(arg)
{
	if not($arg)
		echo count(toedit)
	else
		def iter cnt;
		foreach iter toedit
			if match($arg $iter)
				incr cnt 1
			endif
		loop
		echo $cnt
	endif
}

cmd_goto(arg)
{
	if not(arg)
		set arg = readline("Go to:")
		if not($arg) return endif
	endif
	if not(check(toedit $arg))
		echo "No such item"
		return
	endif
	def cur;
	set cur_item = $arg
	sto.break(toedit handler)
	while;
		set cur = sto.enum(toedit handler)
	loop(neq($cur $cur_item));
}

cmd_next
{
	set cur_item = sto.enum(toedit handler)
	if not($handler) 
		echo "Last item reached, wrapping" 
		set cur_item = sto.enum(toedit handler)
	endif
}

cmd_prev
{
	put toedit __reversed__ true
	cmd_next()
	put toedit __reversed__ ''
}

cmd_change(arg) 
{
	if not($arg)
		set arg = readline("Value for $cur_item\:");
		if not($arg) echo "Cancel" return endif
	endif
	put toedit $cur_item $arg 
}

cmd_delete(arg)
{
	if eqi(readline("$cur_item\: Are you sure\(Y/N)?") y)
		sto.break(toedit handler)
		remove toedit $cur_item
		set cur_item = sto.enum(toedit handler)
	endif
}

cmd_empty
{
	if not($no-confirm)
		if neqi(readline("$cur_item\: Are you sure\(Y/N)?") y)) return endif
	endif
	put toedit $cur_item ''
}

cmd_insert
{
	def new = readline('New key:')
		val = readline('Value:');
	if check(toedit $new)
		echo "Key $new already in table"
		return
	endif
	insert toedit $new $val
}

cmd_copy(arg)
{
	if not($arg)
		set arg = readline('Destination table:')
	endif
	local table(tmp $arg)
	def iter;
	foreach iter toedit
		if check(tmp $iter)
			echo "$iter already in $arg; skipping"
		else
			insert tmp $iter get(toedit $iter)
		endif
	loop
}

cmd_context(arg)
{
	if not($arg) set arg = 1 endif
	def cnt;
	put toedit __reversed__ true
	set cnt = 1
	def tmp;
	while le(int $cnt $arg)
		set tmp = sto.enum(toedit handler)
		echo "-$cnt\: $tmp\: get(toedit $tmp)"
		incr cnt 1
	loop($handler)
	set cnt = 1
	put toedit __reversed__ ''
	while neq($tmp $cur_item)
		set tmp = sto.enum(toedit handler)
	loop
	echo "0: $cur_item\: get(toedit $cur_item)"
	while le(int $cnt $arg)
		set tmp = sto.enum(toedit handler)
		echo "$cnt\: $tmp\: get(toedit $tmp)"
		incr cnt 1
	loop($handler)
	put toedit __reversed__ true
	while neq($tmp $cur_item)
		set tmp = sto.enum(toedit handler)
	loop
	put toedit __reversed__ ''
}

cmd_forall(arg)
{
	if not($arg) return endif
	def cmd;
	set arg = split($arg ' ' '' cmd)
	if not(isproc(cmd_$cmd))
		echo "Invalid command $cmd"
	else
		set no-confirm = true
		def old = $cur_item;
		foreach cur_item toedit
			call cmd_$cmd $arg;
		loop
		set cur_item = $old
		set no-confirm = ''
	endif
}
