#! /usr/local/bin/sl
#
# 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
.ifhas ::module::readline
.module <readline>
.endif
.include <getopt.sci>

main
{
	global tname srcname ifs continuation append quiet override xmlmode; 
	getopt(list(m = =multiline multiline 
				a = =append append 
				i = =ifs ifs = : 
				source = :
				help
				o = override override
				q = quiet quiet
				xml = :
				) process-options)
	if not($tname)
		set tname = trim(readline('Table:'))
		if not($tname) exit endif
	endif
	table tbl cond($append '' !)$tname
	if not(instream())
		if not($srcname)
			set srcname = trim(readline('Input file:'))
		endif
		if not($srcname) 
			if $xmlmode error "An XML file cannot be entered from console" endif
			fill-from-console()
			exit
		endif
		defstream mystream file://cond(eq($srcname -) '' $srcname) cond($xmlmode xml basic);
		instream mystream
	endif
	if $xmlmode
		def curtag bodyval attrs name val;
		global baseval;
		while;
			read
			if iseof() break endif
			if or(eq($[] !--) eq($[] !) eq($[] ?)) continue endif
			if not($[])
				if $curtag
					set bodyval = $bodyval$[value]
				elsif neq($[value] ' ')
					error "Invalid character data in XML file"
				endif
			elsif and($curtag eqi($[] /$curtag))
				xml-put-into-table(get(xmldesc $curtag) $bodyval)
				clear curtag
				clear bodyval
				ioctl instream() nokeepspaces
			else
				set attrs = $[*];
				while $attrs
					set attrs = split($attrs \t \n name val)
					if eq($name /) continue endif
					if not(check(xmldesc $[]::$name))
						error "Unknown attribute $name in tag $[]"
					else
						if check(xmldesc $[]::$name::based)
							set val = $baseval$val
						endif
						xml-put-into-table(get(xmldesc $[]::$name) $val)
					endif
				loop
				if check(xmldesc $[])
					if not(tag(/)) 
						set curtag = $[]
						ioctl instream() keepspaces
						if check(xmldesc $[]::::based)
							set bodyval = $baseval
						endif
					else
						if check(xmldesc $[]::::based)
							xml-put-into-table(get(xmldesc $[]) $baseval)
						endif
					endif
				endif
			endif
		loop;
		def i;
		foreach i xmldesc
			if isprefix($i ::required)
				set i = substr($i sub(length(i) length(::required)))
				if not(check(tname get(xmldesc $i)))
					error "$i is missing"
				endif
			elsif isprefix($i ::default)
				set val = get(xmldesc $i)
				set i = substr($i sub(length(i) length(::default)))
				if not(check(tname get(xmldesc $i)))
					insert tname $i $val
				endif
			endif
		loop
	else
		if not($continuation) defautoseq rl linefilter endif
		if not($ifs) set ifs = \t endif
		def line key val;
		while;
			set line = read(rl)
			if $line
				set val = split($line $ifs '' key)
				put-into-table($key $val)
			endif
		loop(not(iseof()));
	endif
}

process-options(type opt arg extra)
{
	switch $type
		when eq($^ name)
			if not($tname)
				set tname = $opt;
			elsif($srcname)
				error "Too many filenames"
			else
				set srcname = $opt
			endif
			breaksw
		when eq($^ option)
			switch $opt
				when eq($^ multiline)
					set continuation = true
					defautoseq rl linefilter:WithCont
					breaksw
				when eq($^ ifs)
					if isint($arg)
						set ifs = char($arg)
					else
						set ifs = $arg
					endif
					breaksw
				when eq($^ append)
					set append = true
					breaksw
				when eq($^ quiet)
					set quiet = true
					breaksw
				when eq($^ override)
					set override = true
					breaksw
				when eq($^ xml)
					table xmldesc ''
					read-xml-desc($arg)
					set xmlmode = true
					breaksw
				when eq($^ source)
					if $srcname error "Too many filenames" endif
					set srcname = $arg
					breaksw
				when eq($^ help)
echo "Usage: mktable [--multiline] [--ifs=sep-char-or-code] [--append] table-name [[--source] source-file]"
					exit
			endsw
			breaksw
		when true
			error "Unknown option $opt"
	endsw
}

xml-put-into-table(key val)
{
	if eq($key '#BASE')
		set baseval = $val
	elsif neq($key '#IGNORE')
		put-into-table($key $val)
	endif
}

put-into-table(key val)
{
	if check(tbl $key)
		if $override
			put tbl $key $val
		endif
		if not($quiet)
			echo "Duplicate key \'$key\', cond($override replaced ignored)"
		endif
	else
		insert tbl $key $val
	endif
}

fill-from-console()
{
	def inbuf val key;
	if not($ifs)
		set ifs = readline('Field separator:');
	endif
	if not($ifs) set ifs = ' ' endif
	while set(inbuf enh-readline())
		set val = split($inbuf $ifs '' key);
		put-into-table($key $val)
	loop
}

enh-readline()
{
	def line = readline(?);
	if $continuation
		while issuffix($line \\)
			set line = substr($line 0 sub(length($line) 1))readline(>)
		loop
	endif
	return($line)
}


read-xml-desc(name)
{
	defstream xmldesc file://$name xml;
	def curtag;
	while;
		readfrom xmldesc;
		if iseof() break endif
		switch $[]
			when eqi($^ tag)
				if $curtag error "cannot nest TAGs" endif
				if not(tag(NAME)) error "TAG NAME is missing" endif
				if not(tag(/)) set curtag = $[NAME] endif;
				if tag(BODYVAR)
					insert xmldesc $curtag $[BODYVAR]
					if tag(BASED)
						insert xmldesc $curtag::::based ''
					endif
				endif
				breaksw
			when eqi($^ /tag)
				if not($curtag) error "/TAG without TAG" endif
				clear curtag
				breaksw
			when eqi($^ attr)
				if not($curtag) error "ATTR without TAG" endif
				if not(tag(NAME)) error "ATTR NAME is missing" endif
				if tag(DEFAULT)
					insert xmldesc $(curtag)::$[NAME]::default tag(DEFAULT)
				elsif tag(REQUIRED)
					insert xmldesc $(curtag)::$[NAME]::required ''
				endif
				if not(tag(VAR)) error "ATTR VAR is missing" endif
				insert xmldesc $(curtag)::$[NAME] $[VAR]
				if tag(BASED)
					insert xmldesc $(curtag)::$[NAME]::based ''
				endif
				breaksw
			when eq($^ !)
			when eq($^ ?)
			when eq($^ !--)
				# skip comment
				breaksw
			when not($^)
				if eq($[value] ' ') breaksw
				else error "character data not allowed in description"
				endif
			when true
				error "unknown tag $[]"
		endsw
	loop;
	rmstream xmldesc
}